/*
** Module   : P2JIndex.java
** Abstract : Index helper class; represents a Progress or SQL index
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description---------------------
** 001 ECF 20051010   @23180 Created initial version. Describes a database
**                           index and contains several helper methods to
**                           detect index redundancy and sets of minimal
**                           unique index component sets.
** 002 ECF 20060129   @24284 Changes needed for temp table support. Added
**                           convenience constructor and made table name a
**                           settable field after construction.
** 003 ECF 20070424   @33213 Fixed equals and hashCode implementations.
**                           Neither took index component order into
**                           account.
** 004 ECF 20080311   @37478 Added isChar parameter to addComponent().
**                           Also integrated generics.
** 005 ECF 20090428   @42126 Exposed setUnique(boolean) as package private.
**                           Needed by DirtyTempTableHelper.
** 006 VMN 20130830          Added method getOrderByClause().
** 007 VMN 20131126          Fixed getOrderByClause(), removed todo.
** 008 SVL 20140106          Minor change in toString.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import com.goldencode.p2j.persist.dialect.*;

/**
 * A descriptor of an index, which may be either a Progress index or an
 * SQL database index. The latter's features for our purposes here represent
 * a subset of the former's features, so the same class is useful for both.
 * <p>
 * An index is described at two levels:
 * <ol>
 *   <li>attributes of the index itself:
 *   <ul>
 *     <li>table name
 *     <li>index name
 *     <li>unique flag
 *     <li>primary flag
 *     <li>word flag
 *   </ul>
 *   <li>components of the index (1 or more):
 *   <ul>
 *     <li>field or column name
 *     <li>sort direction
 *     <li>abbreviation support
 *     <li>whether data type is character
 *     <li>whether case should be ignored for a character component
 *   </ul>
 * </ol>
 * <p>
 * This class contains several static convenience methods:
 * <ul>
 *   <li>{@link #normalizeUniqueIndexes}: Given a list of indexes for the
 *       same table, this method identifies the minimal set of unique indexes
 *       which define the least common denominator, unique indexes for a
 *       table. It also normalizes any index which contains all of the
 *       components of any of these by marking it as unique.
 *   <li>{@link #nonRedundantIndexes}:  Given a list of indexes for the same
 *       table, this method identifies the sub-list of indexes which are not
 *       redundant with one another. For our purposes here, an index is
 *       considered redundant with another, if both indexes have the same
 *       number of components, where each component appears in the same order
 *       and has the same name as its counterpart.
 * </ul>
 */
public final class P2JIndex
{
   /** Name of index */
   private final String name;
   
   /** Is this the primary index for its table? */
   private final boolean primary;
   
   /** Is this a word index? */
   private final boolean word;
   
   /** List of index components */
   private final List<P2JIndexComponent> components =
      new ArrayList<P2JIndexComponent>();
   
   /** Name of table with which index is associated */
   private String table;
   
   /** Does this index define a unique constraint for its table? */
   private boolean unique;
   
   /**
    * Convenience constructor.
    *
    * @param   table
    *          Name of the table with which the index is associated.
    * @param   name
    *          Index name.
    * @param   unique
    *          <code>true</code> if this index defines a unique constraint
    *          for its table, else <code>false</code>.
    */
   public P2JIndex(String table, String name, boolean unique)
   {
      this(table, name, unique, false, false);
   }
   
   /**
    * Convenience constructor.
    *
    * @param   table
    *          Name of the table with which the index is associated.
    * @param   name
    *          Index name.
    * @param   unique
    *          <code>true</code> if this index defines a unique constraint
    *          for its table, else <code>false</code>.
    * @param   primary
    *          <code>true</code> if this index is the primary index for its
    *          table, else <code>false</code>.
    * @param   word
    *          <code>true</code> if this index is a word index, else
    *          <code>false</code>.
    */
   public P2JIndex(String table,
                   String name,
                   boolean unique,
                   boolean primary,
                   boolean word)
   {
      this.table = table;
      this.name = name;
      this.unique = unique;
      this.primary = primary;
      this.word = word;
   }
   
   /**
    * Given a list of indexes, identify the minimal set of unique indexes
    * which define the least common denominator, unique constraints for a
    * table. This will be the set of unique indexes which are least specific
    * in terms of components. For instance, given three unique indexes with
    * the following components:
    * <ul>
    *   <li>Index 1:  A
    *   <li>Index 2:  A, B
    *   <li>Index 3:  A, B, C
    * </ul>
    * Index 1 is the least common denominator, unique index of the group.
    * This is because the table can contain no tuples where the value of
    * columns B or C affect the uniqueness of the tuple. That is, for every
    * unique value of A, there can be only one combination of B and C, since
    * no other tuple can exist with the same value of A but different values
    * of B or C.
    * <p>
    * Conversely, if Index 2 and Index 3 are not explicitly defined as being
    * unique in a Progress schema, they are nevertheless <i>implicitly</i>
    * unique because they include column A, which requires each tuple to
    * contain a unique value for column A. Thus, as a useful side effect,
    * this method normalizes such indexes by marking them unique, regardless
    * of whether the unique flag was applied in the originating Progress
    * schema.
    *
    * @param   indexes
    *          A list of indexes for a common table.
    *
    * @return  The list of indexes which defines the minimal set of least
    *          common denominator, unique indexes for that index group.
    *          Order of the original list is not guaranteed to be preserved
    *          in the resulting subset list.
    *
    * @throws  IllegalArgumentException
    *          if <code>indexes</code> contains index descriptors for
    *          different tables.
    */
   public static List<P2JIndex> normalizeUniqueIndexes(List<P2JIndex> indexes)
   {
      Map<Set<String>, P2JIndex> results =
         new HashMap<Set<String>, P2JIndex>();
      int size = indexes.size();
      
      // Iterate backwards through the full list to get the next index to
      // test as a least common denominator candidate.
      ListIterator<P2JIndex> iIter = indexes.listIterator(size);
      while (iIter.hasPrevious())
      {
         P2JIndex iIndex = iIter.previous();
         boolean minimal = iIndex.isUnique();
         
         // Iterate backwards through the full list to test each other index
         // in the list against the LCD candidate.
         ListIterator<P2JIndex> jIter = indexes.listIterator(size);
         while (jIter.hasPrevious())
         {
            P2JIndex jIndex = jIter.previous();
            
            // Skip the same descriptor instance.
            if (iIndex == jIndex)
            {
               continue;
            }
            
            // Assert that both test subjects are from the same table.
            if (!iIndex.getTable().equals(jIndex.getTable()))
            {
               throw new IllegalArgumentException(
                  "Cannot normalize indexes from different tables;  found '" +
                  iIndex + "' and '" + jIndex);
            }
            
            // Is the LCD candidate a possible superset of another index?
            if (iIndex.containsAll(jIndex))
            {
               // The normalization step:  if the LCD candidate contains all
               // components of a unique index, it is itself unique, by
               // definition.
               if (jIndex.isUnique())
               {
                  iIndex.setUnique(true);
                  minimal = true;
               }
               
               // Candidate cannot be confirmed if:
               // A) its components form a superset of another index'
               //    components;  OR
               // B) neither it nor the counterpart index is unique.
               int iSize = iIndex.size();
               int jSize = jIndex.size();
               if ((iSize > jSize && jIndex.isUnique()) ||
                   (!minimal && !jIndex.isUnique()))
               {
                  minimal = false;
               }
               
               // No need for further consideration if we've already ruled
               // out LCDness.
               if (!minimal)
               {
                  break;
               }
            }
         }
         
         // If we got through both loop levels and the candidate was
         // confirmed, store it in a map, keyed by the candidate's name set.
         if (minimal)
         {
            results.put(iIndex.getNameSet(null), iIndex);
         }
      }
      
      return (new ArrayList<P2JIndex>(results.values()));
   }
   
   /**
    * Given a list of index descriptors, assemble a list of those which are
    * not redundant with one another.  An index is considered redundant if it
    * contains all components of another, and in the same order.  For every
    * group of redundant indexes, one (the first encountered in a first to
    * last scan) is added to the resulting, non-redundant list.
    *
    * @param   indexes
    *          List of indexes which contain possible redundancies.
    *
    * @return  Sublist of <code>indexes</code> which contains no redundancies.
    */
   public static List<P2JIndex> nonRedundantIndexes(List<P2JIndex> indexes)
   {
      List<P2JIndex> list = new ArrayList<P2JIndex>();
      int size = indexes.size();
      for (int i = 0; i < size; i++)
      {
         P2JIndex iIndex = indexes.get(i);
         
         boolean redundant = false;
         
         for (int j = 0; j < size; j++)
         {
            if (j == i)
            {
               continue;
            }
            
            P2JIndex jIndex = indexes.get(j);
            int iSize = iIndex.size();
            int jSize = jIndex.size();
            
            if (iSize == jSize && iIndex.leadsWith(jIndex) && i > j)
            {
               redundant = true;
               break;
            }
         }
         
         if (!redundant)
         {
            list.add(iIndex);
         }
      }
      
      return list;
   }
   
   /**
    * Get table name associated with this index.
    *
    * @return  Table name.
    */
   public String getTable()
   {
      return table;
   }
   
   /**
    * Get name of this index.
    *
    * @return  Index name.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Indicate whether this index is unique.
    *
    * @return  <code>true</code> if unique, else <code>false</code>.
    */
   public boolean isUnique()
   {
      return unique;
   }
   
   /**
    * Indicate whether this index is the primary index for its table (only
    * meaningful for Progress indexes).
    *
    * @return  <code>true</code> if primary, else <code>false</code>.
    */
   public boolean isPrimary()
   {
      return primary;
   }
   
   /**
    * Indicate whether this index is a word index (only meaningful for
    * Progress indexes).
    *
    * @return  <code>true</code> if a word index, else <code>false</code>.
    */
   public boolean isWord()
   {
      return word;
   }
   
   /**
    * Add an index component to the index definition. An index component
    * consists of a field name and flags indicating sort direction, whether
    * the component is a character data type, case sensitivity, and
    * abbreviation. This method variant sets the descending flag, the
    * character flag, the caseSensitive flag, and the abbreviated flag to
    * <code>false</code>.
    *
    * @param   name
    *          Name of the index field which defines this index component.
    * 
    * @return  Newly added <code>P2JIndexComponent</code> object.
    */
   public P2JIndexComponent addComponent(String name)
   {
      return addComponent(name, false, false, false, false);
   }
   
   /**
    * Add an index component to the index definition. An index component
    * consists of a field name, and a set of flags indicating sort direction,
    * whether the component is a character data type, case sensitivity, and
    * abbreviation. This method variant sets both the descending and
    * abbreviated flags to <code>false</code>, and assumes that the component
    * is a character type, since <code>ignoreCase</code> can go either way.
    *
    * @param   name
    *          Name of the index field which defines this index component.
    * @param   ignoreCase
    *          <code>true</code> if this component represents a character
    *          field AND case is ignored when sorting index, else
    *          <code>false</code>.
    * 
    * @return  Newly added <code>P2JIndexComponent</code> object.
    */
   public P2JIndexComponent addComponent(String name, boolean ignoreCase)
   {
      return addComponent(name, false, true, ignoreCase, false);
   }
   
   /**
    * Add an index component to the index definition. An index component
    * consists of a field name, sort direction, case sensitivity flag, and
    * abbreviation flag.
    *
    * @param   name
    *          Name of the index field which defines this index component.
    * @param   descending
    *          <code>true</code> if this component sorts naturally in
    *          descending order; <code>false</code> to sort in ascending
    *          order.
    * @param   isChar
    *          <code>true</code> if this component represents a character
    *          field.
    * @param   ignoreCase
    *          <code>true</code> if this component represents a character
    *          field AND case is ignored when sorting index, else
    *          <code>false</code>.
    * @param   abbreviated
    *          <code>true</code> if this component is abbreviated, else
    *          <code>false</code>.
    * 
    * @return  Newly added <code>P2JIndexComponent</code> object.
    */
   public P2JIndexComponent addComponent(String name,
                                         boolean descending,
                                         boolean isChar,
                                         boolean ignoreCase,
                                         boolean abbreviated)
   {
      P2JIndexComponent comp = new P2JIndexComponent(name,
                                                     descending,
                                                     isChar,
                                                     ignoreCase,
                                                     abbreviated);
      components.add(comp);
      
      return comp;
   }
   
   /**
    * Get the number of components in this index.
    *
    * @return  Index size.
    */
   public int size()
   {
      return components.size();
   }
   
   /**
    * Get the index component at the specified position.
    *
    * @param   position
    *          0-based position of the desired index component.
    *
    * @return  Index component at <code>position</code>.
    *
    * @throws  ArrayIndexOutOfBoundsException
    *          if <code>position</code> is invalid.
    */
   public P2JIndexComponent getComponent(int position)
   {
      return components.get(position);
   }
   
   /**
    * Get an iterator on all {@link P2JIndexComponent}s in this index.
    *
    * @return  Component iterator.
    */
   public Iterator<P2JIndexComponent> components()
   {
      return components.iterator();
   }
   
   /**
    * Return up to <code>size</code> leading components of this index. If
    * <code>size</code> is larger than the number of components in the index,
    * all components are returned. The order of components is preserved in
    * the returned list.
    *
    * @param   size
    *          Number of leading components to return.
    *
    * @return  List view of the first <code>size</code> components of the
    *          index.
    */
   public List<P2JIndexComponent> getLeadingComponents(int size)
   {
      size = Math.min(size, components.size());
      
      return components.subList(0, size);
   }

   /**
    * Get order by clause for given index of database.
    *
    * @param   database
    *          Database.
    * @param   dmoAlias
    *          Alias qualifier which represents DMO in sort phrase.
    * @param   dmoClass
    *          DMO implementation class.
    *
    * @return  order by clause.
    */
   public String getOrderByClause(Database database, String dmoAlias, Class dmoClass)
   {
      StringBuilder result = new StringBuilder();
      Iterator comps = components();
      if (comps.hasNext())
      {
         Map<String, String> mappedCols = DatabaseManager.getColumnToPropertyMap(database,
                                                                                 dmoClass,
                                                                                 null,
                                                                                 true);
         P2JIndexComponent nextComp;
         while (comps.hasNext())
         {
            if (result.length() > 0)
            {
               result.append(", ");
            }
            nextComp = (P2JIndexComponent) comps.next();
            String field = mappedCols.get(nextComp.getName());

            result.append(dmoAlias);
            result.append(".");
            result.append(field);
            result.append(" ");
            result.append(nextComp.isDescending() ? "desc": "asc");
         }
      }
      return result.toString();
   }

   /**
    * Test whether this index leads with components of the same name as the
    * specified index. For the purpose of this test, only the names of the
    * components are considered;  the sort direction and word flags are
    * ignored. The order of the index components is significant.
    *
    * @param   index
    *          Index with which to compare this index' components.
    *
    * @return  <code>true</code> if and only if this index contains <b>all</b>
    *          of the components of <code>index</code> as its leading
    *          components, and these components appear in the same order in
    *          both indexes; <code>false</code> otherwise.
    */
   public boolean leadsWith(P2JIndex index)
   {
      int size = index.size();
      if (size > components.size())
      {
         return false;
      }
      
      for (int i = 0; i < size; i++)
      {
         P2JIndexComponent next = index.components.get(i);
         P2JIndexComponent local = components.get(i);
         
         if (!next.getName().equals(local.getName()))
         {
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Return the set of index component names for this index.  Iterations
    * over this set will return the names in the order in which they were
    * added to the index.
    *
    * @param   dialect
    *          Optional P2J database dialect.  If provided, index components
    *          of text type will be processed as necessary to produce an
    *          expression which will right trim whitespace and properly handle
    *          case insensitivity as necessary.  This may mean wrapping the
    *          property name in various functions, substituting the name with
    *          a computed column name, or whatever is appropriate for the
    *          particular dialect.
    *
    * @return  Set of index component names.
    */
   public LinkedHashSet<String> getNameSet(P2JDialect dialect)
   {
      LinkedHashSet<String> set = new LinkedHashSet<String>();
      Iterator<P2JIndexComponent> iter = components.iterator();
      while (iter.hasNext())
      {
         P2JIndexComponent next = iter.next();
         String name = next.getName();
         if (dialect != null && next.isCharType())
         {
            boolean ignoreCase = next.isIgnoreCase();
            name = dialect.getProcessedCharacterColumnName(name, ignoreCase);
         }
         set.add(name);
      }
      
      return set;
   }
   
   /**
    * Implement a very loose test for equivalence, whereby <code>o</code> is
    * considered equivalent to this object as long as both instances contain
    * exactly the same set of component names, in the same order. Table name,
    * index name, and component sort direction and abbreviation support are
    * ignored for this test.
    *
    * @param   o
    *          An object instance, presumably a <code>P2JIndex</code> object,
    *          to test for equivalence with this object.
    *
    * @return  <code>true</code> if <code>o</code> tests as equivalent to
    *          this object, using the criteria specified above, else
    *          <code>false</code>.
    */
   public boolean equals(Object o)
   {
      if (!(o instanceof P2JIndex))
      {
         return false;
      }
      P2JIndex idx = (P2JIndex) o;
      
      // Can't use LinkedHashSet.equals() to compare name sets returned by
      // getNameSet(), because that equals() implementation does not take
      // element order into account.  Do the comparison manually.
      
      Collection<P2JIndexComponent> c1 = this.components;
      Collection<P2JIndexComponent> c2 = idx.components;
      
      if (c1.size() != c2.size())
      {
         return false;
      }
      
      Iterator<P2JIndexComponent> iter1 = c1.iterator();
      Iterator<P2JIndexComponent> iter2 = c2.iterator();
      while (iter1.hasNext())
      {
         P2JIndexComponent next1 = iter1.next();
         P2JIndexComponent next2 = iter2.next();
         
         if (!next1.getName().equals(next2.getName()))
         {
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Overridden implementation to ensure consistency with our specialized
    * {@link #equals} method implementation.  Note that only the index
    * component names are taken into account;  all other index and index
    * component data is ignored for purposes of this calculation.
    *
    * @return  Hash code of this index.
    */
   public int hashCode()
   {
      int result = 17;
      Iterator<P2JIndexComponent> iter = components.iterator();
      while (iter.hasNext())
      {
         P2JIndexComponent next = iter.next();
         result = 37 * next.getName().hashCode();
      }
      
      return result;
   }
   
   /**
    * Get a string representation of the internal state of this object,
    * primarily for debugging purposes.
    *
    * @return  String representing this object's state.
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder();
      
      if (table != null)
      {
         buf.append(table);
         buf.append('.');
      }
      buf.append(name);
      buf.append(" [");
      
      Iterator<P2JIndexComponent> iter = components.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         P2JIndexComponent next = iter.next();
         if (i > 0)
         {
            buf.append(", ");
         }
         buf.append(next.getName());
      }
      
      buf.append(']');
      
      if (isUnique())
      {
         buf.append('*');
      }
      
      if (isPrimary())
      {
         buf.append('!');
      }
      
      if (isWord())
      {
         buf.append('%');
      }
      
      return buf.toString();
   }
   
   /**
    * Set the table with which this index is associated.
    *
    * @param   table
    *          Name of table.
    */
   void setTable(String table)
   {
      this.table = table;
   }
   
   /**
    * Set the unique flag within this index.
    *
    * @param   unique
    *          <code>true</code> to indicate that this index is unique;
    *          <code>false</code> to indicate it is not.
    */
   void setUnique(boolean unique)
   {
      this.unique = unique;
   }
   
   /**
    * Indicate whether this index contains all of the components of the
    * specified index.  Only index component names are considered, not any
    * other properties of each component.
    *
    * @param   index
    *          Index whose components are checked against those in this
    *          index.
    *
    * @return  <code>true</code> if this object contains all components of
    *          <code>index</code>;  else <code>false</code>.
    */
   private boolean containsAll(P2JIndex index)
   {
      Set<String> localSet = getNameSet(null);
      Set<String> otherSet = index.getNameSet(null);
      
      return localSet.containsAll(otherSet);
   }
   
   /**
    * Test harness. Creates multiple instance of this class and tests that
    * minimal unique indexes can be determined, and that redundant indexes
    * can be identified and culled from a list.
    *
    * @param   args
    *          Not used.
    */
   public static void main(String[] args)
   {
      try
      {
         P2JIndex index;
         List<P2JIndex> list = new ArrayList<P2JIndex>();
         
         index = new P2JIndex("Table 1", "ABCD", true, false, false);
         index.addComponent("A");
         index.addComponent("B");
         index.addComponent("C");
         index.addComponent("D");
         list.add(index);
         
         index = new P2JIndex("Table 1", "ACB", true, false, false);
         index.addComponent("A");
         index.addComponent("C");
         index.addComponent("B");
         list.add(index);
         
         index = new P2JIndex("Table 1", "BCA", true, false, false);
         index.addComponent("B");
         index.addComponent("C");
         index.addComponent("A");
         list.add(index);
         
         index = new P2JIndex("Table 1", "DE", true, false, false);
         index.addComponent("D");
         index.addComponent("E");
         list.add(index);
         
         index = new P2JIndex("Table 1", "BAC", true, false, false);
         index.addComponent("B");
         index.addComponent("A");
         index.addComponent("C");
         list.add(index);
         
         index = new P2JIndex("Table 1", "BC", true, false, false);
         index.addComponent("B");
         index.addComponent("C");
         list.add(index);
         
         index = new P2JIndex("Table 1", "AB", true, false, false);
         index.addComponent("A");
         index.addComponent("B");
         list.add(index);
         
         index = new P2JIndex("Table 1", "F1", false, false, false);
         index.addComponent("F");
         list.add(index);
         
         index = new P2JIndex("Table 1", "F2", true, true, false);
         index.addComponent("F");
         list.add(index);
         
         index = new P2JIndex("Table 1", "ED1", false, false, false);
         index.addComponent("E");
         index.addComponent("D");
         list.add(index);
         
         index = new P2JIndex("Table 1", "ED2", false, false, false);
         index.addComponent("E");
         index.addComponent("D");
         list.add(index);
         
         index = new P2JIndex("Table 1", "ED3", true, false, false);
         index.addComponent("E");
         index.addComponent("D");
         list.add(index);
         
         index = new P2JIndex("Table 1", "EG", false, false, false);
         index.addComponent("E");
         index.addComponent("G", true);
         list.add(index);
         
         index = new P2JIndex("Table 1", "CAB", true, false, false);
         index.addComponent("C");
         index.addComponent("A");
         index.addComponent("B");
         list.add(index);
         
         System.out.println("PRE-NORMALIZED (" + list.size() + ")");
         System.out.println("-------------------------------");
         System.out.println(list);
         System.out.println();
         
         List<P2JIndex> minimalUnique = P2JIndex.normalizeUniqueIndexes(list);
         List<P2JIndex> nonRedundant = P2JIndex.nonRedundantIndexes(list);
         
         System.out.println("NON-REDUNDANT (" + nonRedundant.size() + ")");
         System.out.println("-------------------------------");
         System.out.println(nonRedundant);
         System.out.println();
         
         System.out.println("MINIMAL UNIQUE (" + minimalUnique.size() + ")");
         System.out.println("-------------------------------");
         System.out.println(minimalUnique);
         System.out.println();
         
         Set<P2JIndex> remaining = new HashSet<P2JIndex>(nonRedundant);
         remaining.removeAll(minimalUnique);
         System.out.println("REMAINING (" + remaining.size() + ")");
         System.out.println("-------------------------------");
         System.out.println(remaining);
         System.out.println();
         
         System.out.println("NORMALIZED (" + list.size() + ")");
         System.out.println("-------------------------------");
         System.out.println(list);
         System.out.println();
      }
      catch (Exception exc)
      {
         exc.printStackTrace();
      }
   }
}
